# vim:ts=8:sts=4:sw=4
# -*- coding: utf-8 -*-

import gobject
import gtk
import xml.sax.saxutils
import gettext

_ = gettext.gettext

TILE_FILE = 'rummikub.png'
TILE_WIDTH = 32
TILE_HEIGHT = 44
TILE_PIXBUF = gtk.gdk.pixbuf_new_from_file(TILE_FILE)

def get_tile_list ():
    return range(52) + range(52) + [52, 53]

def get_tile_tuple (tile):
    return tile / 13, tile % 13

def create_board (widget_name, bgcolor, cols, rows):
    board = Board(TILE_PIXBUF, bgcolor, cols, rows)
    board.set_size_request(cols * TILE_WIDTH, rows * TILE_HEIGHT)
    board.show()
    return board

class CellRendererTiles (gtk.GenericCellRenderer):
    buffer = None
    __gproperties__ = {
	'tiles': (gobject.TYPE_OBJECT, 'tiles', 'tiles',
		gobject.PARAM_READWRITE),
    }

    def __init__ (self, image, bgcolor):
	gtk.GenericCellRenderer.__init__(self)
	self.IMAGE = image
	self.BGCOLOR = gtk.gdk.color_parse(bgcolor)
	self.tiles = None

    def do_set_property (self, pspec, value):
	setattr(self, pspec.name, value)

    def do_get_property (self, pspec):
	return getattr(self, pspec.name)

    def on_render (self, window, widget,
	    background_area, cell_area, expose_area, flags):
	if self.tiles == None:
	    return
	tiles = self.tiles.tiles
	cols = len(tiles)
	width = TILE_WIDTH * cols
	height = TILE_HEIGHT

	if cols == 0:
	    return
	if self.tiles.buffer == None:
	    image_width = width * 55. / cols
	    image_height = height * 2.
	    self.scaled_image = self.IMAGE.scale_simple(
		    int(image_width), int(image_height), gtk.gdk.INTERP_TILES)

	    self.tiles.buffer = gtk.gdk.Pixmap(widget.window, width, height, -1)
	    if self.tiles.gc == None:
		self.tiles.gc = widget.window.new_gc()
		self.tiles.gc.copy(widget.style.black_gc)
		self.tiles.gc.set_rgb_fg_color(self.BGCOLOR)
	    self.tiles.buffer.draw_rectangle(self.tiles.gc, True,
		    0, 0, width, height)
	    for c in range(cols):
		self.tiles.buffer.draw_pixbuf(self.tiles.gc, self.scaled_image,
			tiles[c] * TILE_WIDTH, 0,
			c * TILE_WIDTH, 0, TILE_WIDTH, TILE_HEIGHT,
			gtk.gdk.RGB_DITHER_NONE, 0, 0)

	buffer_width, buffer_height = self.tiles.buffer.get_size()
	window.draw_drawable(self.tiles.gc, self.tiles.buffer, 0, 0,
		cell_area.x, cell_area.y, buffer_width, buffer_height)

    def on_get_size (self, widget, cell_area):
	if self.tiles == None:
	    return 0, 0, 0, 0
	cols = len(self.tiles.tiles)
	if cols == 0:
	    return 0, 0, 0, 0
	width = TILE_WIDTH * cols
	height = TILE_HEIGHT

	return 0, 0, width, height

gobject.type_register(CellRendererTiles)

class Tiles (gobject.GObject):
    buffer = None
    gc = None

    def __init__ (self, tiles):
	gobject.GObject.__init__(self)
	self.tiles = tiles[:]

gobject.type_register(Tiles)

class BoardException (Exception):
    pass

class Board (gtk.DrawingArea):
    buffer = None
    selected = None
    gc = None

    def __init__ (self, image, bgcolor, cols, rows):
	gtk.DrawingArea.__init__(self)
	self.add_events(gtk.gdk.BUTTON_PRESS_MASK)
	self.IMAGE = image
	self.BGCOLOR = gtk.gdk.color_parse(bgcolor)
	self.COLS = cols
	self.ROWS = rows
	self.tiles = []
	for i in range(self.ROWS):
	    r = []
	    for j in range(self.COLS):
		r.append(-1)
	    self.tiles.append(r)

    def do_set_property(self, pspec, value):
	setattr(self, pspec.name, value)

    def do_get_property(self, pspec):
	return getattr(self, pspec.name)

    def tile_coord (self, x, y):
	buffer_width, buffer_height = self.buffer.get_size()
	row = int(y * self.ROWS / buffer_height)
	col = int(x * self.COLS / buffer_width)
	if row >= self.ROWS or col >= self.COLS:
	    raise BoardException
	return row, col

    def selected_tile (self, row, col):
	return self.tiles[row][col]

    def get_tile (self, row, col):
	return self.tiles[row][col]

    def get_all_tile (self):
	tiles = []
	for r in self.tiles[:]:
	    tiles.append(r[:])
	return tiles

    def set_all_tiles (self, tiles):
	self.tiles = tiles
	self.draw_all_tile()

    def reset_all_tiles (self):
	for r in range(self.ROWS):
	    for c in range(self.COLS):
		self.tiles[r][c] = -1
	self.draw_all_tile()

    def select_tile (self, row, col):
	if self.selected:
	    raise BoardException
	if self.tiles[row][col] == -1:
	    raise BoardException
	self.selected = row, col
	self.draw_tile(row, col)

    def unselect_tile (self):
	if not self.selected:
	    raise BoardException
	row, col = self.selected
	self.selected = None
	self.draw_tile(row, col)

    def add_tile (self, tile):
	for r in range(self.ROWS):
	    for c in range(self.COLS):
		if self.tiles[r][c] == -1:
		    self.tiles[r][c] = tile
		    self.draw_tile(r, c)
		    return
	raise BoardException

    def insert_tile (self, tile, row, col):
	if self.tiles[row][col] != -1:
	    raise BoardException
	self.tiles[row][col] = tile
	self.draw_tile(row, col)

    def remove_tile (self, row, col):
	if self.tiles[row][col] == -1:
	    raise BoardException
	self.tiles[row][col] = -1
	self.draw_tile(row, col)

    def move_tile (self, frow, fcol, row, col):
	if self.tiles[frow][fcol] == -1:
	    raise BoardException
	if self.tiles[row][col] != -1:
	    raise BoardException
	self.tiles[row][col] = self.tiles[frow][fcol]
	self.tiles[frow][fcol] = -1
	self.draw_tile(frow, fcol)
	self.draw_tile(row, col)

    def exchange_tile (self, frow, fcol, row, col):
	if self.tiles[frow][fcol] == -1:
	    raise BoardException
	if self.tiles[row][col] == -1:
	    raise BoardException
	tile = self.tiles[row][col]
	self.tiles[row][col] = self.tiles[frow][fcol]
	self.tiles[frow][fcol] = tile
	self.draw_tile(frow, fcol)
	self.draw_tile(row, col)

    def draw_all_tile (self):
	buffer_width, buffer_height = self.buffer.get_size()
	self.buffer.draw_rectangle(self.gc, True, 0, 0,
		buffer_width, buffer_height)

	for r in range(self.ROWS):
	    for c in range(self.COLS):
		if self.tiles[r][c] == -1:
		    continue
		dest_x = buffer_width * c / self.COLS
		dest_y = buffer_height * r / self.ROWS

		width = int(buffer_width / self.COLS)
		height = int(buffer_height / self.ROWS)

		tile = self.tiles[r][c]
		x = int(buffer_width * tile / self.COLS)
		if self.selected == (r, c):
		    y = int(buffer_height / self.ROWS)
		else:
		    y = 0

		self.buffer.draw_pixbuf(self.gc, self.scaled_image,
			x, y, dest_x, dest_y, width, height,
			gtk.gdk.RGB_DITHER_NONE, 0, 0)

	self.queue_draw_area(0, 0, buffer_width, buffer_height)

    def draw_tile (self, row, col):
	buffer_width, buffer_height = self.buffer.get_size()
	dest_x = buffer_width * col / self.COLS
	dest_y = buffer_height * row / self.ROWS

	width = int(buffer_width / self.COLS)
	height = int(buffer_height / self.ROWS)

	self.buffer.draw_rectangle(self.gc, True, dest_x, dest_y, width, height)
	if self.tiles[row][col] != -1:
	    tile = self.tiles[row][col]
	    x = int(buffer_width * tile / self.COLS)
	    if self.selected == (row, col):
		y = int(buffer_height / self.ROWS)
	    else:
		y = 0

	    self.buffer.draw_pixbuf(self.gc, self.scaled_image,
		    x, y, dest_x, dest_y, width, height,
		    gtk.gdk.RGB_DITHER_NONE, 0, 0)

	self.queue_draw_area(dest_x, dest_y, width, height)

    def resize (self, width, height):
	image_width = width * 55. / self.COLS
	image_height = height * 2. / self.ROWS
	self.scaled_image = self.IMAGE.scale_simple(
		int(image_width), int(image_height), gtk.gdk.INTERP_TILES)

	self.buffer = gtk.gdk.Pixmap(self.window, width, height, -1)
	if self.gc == None:
	    self.gc = self.window.new_gc()
	    self.gc.copy(self.style.black_gc)
	    self.gc.set_rgb_fg_color(self.BGCOLOR)
	self.buffer.draw_rectangle(self.gc, True, 0, 0, width, height)

	self.draw_all_tile()

    def do_configure_event (self, event):
	width = event.width
	height = event.height
	self.resize (event.width, event.height)

    def do_expose_event (self, event):
	if self.buffer == None:
	    geometry = self.window.get_geometry()
	    self.resize(geometry[2], geometry[3])

	self.window.draw_drawable(self.gc, self.buffer,
		event.area.x, event.area.y, event.area.x, event.area.y,
		event.area.width, event.area.height)

	return True

gobject.type_register(Board)

class Drawing:
    image = gtk.gdk.pixbuf_new_from_file(TILE_FILE)
    gamers = gtk.ListStore(int, str, int, int, Tiles)
    time30 = None
    time50 = None
    time55 = None

    def __init__ (self, xml):
	self.xml = xml
	self.xml.signal_autoconnect(self)

	self.board = self.xml.get_widget('board')
	self.deck = self.xml.get_widget('deck')

	gamer_list = self.xml.get_widget('gamer_list')
	gamer_list.set_model(self.gamers)
	gamer_list.insert_column_with_attributes(
		-1, _('Name'), gtk.CellRendererText(), text=1)
	gamer_list.insert_column_with_attributes(
		-1, _('Score'), gtk.CellRendererText(), text=2)

	score_list = self.xml.get_widget('score_list')
	score_list.set_model(self.gamers)
	score_list.insert_column_with_attributes(
		-1, _('Name'), gtk.CellRendererText(), text=1)
	score_list.insert_column_with_attributes(
		-1, _('Score'), gtk.CellRendererText(), text=2)
	score_list.insert_column_with_attributes(
		-1, _('This Round'), gtk.CellRendererText(), text=3)
	score_list.insert_column_with_attributes(
		-1, _('Remaining Tiles'),
		CellRendererTiles(TILE_PIXBUF, '#ffffff'), tiles=4)

	main_window = self.xml.get_widget('rummikub')
	x, y = main_window.get_position()
	width, height = main_window.get_size()

	self.xml.get_widget('deck_window').move(x, y + height)
	self.xml.get_widget('control_window').move(x + width, y)

    def start_round (self, tiles):
	self.message(_('Rummikub started'))
	self.my_message(_('Rummikub started'))

	self.deck.reset_all_tiles()
	for t in tiles:
	    self.deck.add_tile(t)
	self.board.reset_all_tiles()

    def end_round (self, score, totalscore, remain):
	self.message(_('Rummikub ended'))
	self.my_message(_('Rummikub ended'))
	if self.time30:
	    gobject.source_remove(self.time30)
	if self.time50:
	    gobject.source_remove(self.time50)
	if self.time55:
	    gobject.source_remove(self.time55)

	iter = self.gamers.get_iter_first()
	while iter:
	    current = self.gamers.get_value(iter, 0)
	    self.gamers.set_value(iter, 2, totalscore[current])
	    self.gamers.set_value(iter, 3, score[current])
	    self.gamers.set_value(iter, 4, Tiles(remain[current]))
	    iter = self.gamers.iter_next(iter)

	self.xml.get_widget('score_dialog').run()
	self.xml.get_widget('score_menu').set_sensitive(True)

    def stop_round (self):
	self.message(_('Rummikub stopped'))
	self.my_message(_('Rummikub stopped'))
	if self.time30:
	    gobject.source_remove(self.time30)
	if self.time50:
	    gobject.source_remove(self.time50)
	if self.time55:
	    gobject.source_remove(self.time55)

    def show_deck (self):
	self.xml.get_widget('deck_menu').set_active(True)

    def hide_deck (self):
	self.xml.get_widget('deck_menu').set_active(False)

    def show_control (self):
	self.xml.get_widget('control_menu').set_active(True)
    
    def hide_control (self):
	self.xml.get_widget('control_menu').set_active(False)

    def set_start_game (self, enable):
	self.xml.get_widget('start_game_button').set_sensitive(enable)

    def set_end_game (self, enable):
	self.xml.get_widget('end_game_button').set_sensitive(enable)

    def add_gamer (self, id, name, totalscore):
	self.message(_('%s entered') % name)
	iter = self.gamers.get_iter_first()
	while iter:
	    current = self.gamers.get_value(iter, 0)
	    if current > id:
		self.gamers.insert_before(iter, (id, name, totalscore, 0, None))
		break;
	    iter = self.gamers.iter_next(iter)
	else:
	    self.gamers.append((id, name, totalscore, 0, None))

    def del_gamer (self, id, name):
	self.message(_('%s quited') % name)
	iter = self.gamers.get_iter_first()
	while iter:
	    current = self.gamers.get_value(iter, 0)
	    if current == id:
		self.gamers.remove(iter)
		break;
	    iter = self.gamers.iter_next(iter)

    def end_game (self):
	text_buffer = self.xml.get_widget('chat_text').get_buffer()
	text_buffer.set_text('')
	self.gamers.clear()

	self.board.reset_all_tiles()
	self.deck.reset_all_tiles()

    def start_turn (self, id, name):
	self.message(_('%s\'s turn started') % name)
    
    def register_success (self, id, name):
	self.message(_('%s succeed to register') % name)

    def register_fail (self, id, name):
	self.message(_('%s failed to register') % name)

    def end_register (self):
	self.xml.get_widget('register_button').set_sensitive(False)

    def start_register_turn (self):
	self.xml.get_widget('register_button').set_sensitive(True)
	self.xml.get_widget('finish_turn_button').set_sensitive(True)
	self.my_message(_('You should register'))
	self.time30 = gobject.timeout_add(30000, self.time30_message)
	self.time50 = gobject.timeout_add(50000, self.time50_message)
	self.time55 = gobject.timeout_add(55000, self.time55_message)

    def start_normal_turn (self):
	self.xml.get_widget('finish_turn_button').set_sensitive(True)
	self.my_message(_('It\'s your turn'))
	self.time30 = gobject.timeout_add(30000, self.time30_message)
	self.time50 = gobject.timeout_add(50000, self.time50_message)
	self.time55 = gobject.timeout_add(55000, self.time55_message)
    
    def time30_message (self):
	self.time30 = None
	self.my_message(_('You\'ve spent 30 seconds'))
	return False

    def time50_message (self):
	self.time50 = None
	self.my_message(_('10 seconds remain'))
	return False

    def time55_message (self):
	self.time55 = None
	self.my_message(_('Hurry up, there\'re only 5 seconds'))
	return False

    def finish_register_turn (self):
	self.xml.get_widget('register_button').set_sensitive(False)
	self.xml.get_widget('finish_turn_button').set_sensitive(False)
	self.my_message('')
	if self.time30:
	    gobject.source_remove(self.time30)
	if self.time50:
	    gobject.source_remove(self.time50)
	if self.time55:
	    gobject.source_remove(self.time55)

    def finish_normal_turn (self):
	self.xml.get_widget('finish_turn_button').set_sensitive(False)
	self.my_message('')
	if self.time30:
	    gobject.source_remove(self.time30)
	if self.time50:
	    gobject.source_remove(self.time50)
	if self.time55:
	    gobject.source_remove(self.time55)

    def finish_turn (self, id, name, type):
	if type == 0:
	    self.message(_('%s finished turn') % name)
	elif type == 1:
	    self.message(_('%s gave up turn (gave 1 tile)') % name)
	elif type == 2:
	    self.message(_('%s rollbacked turn (gave 3 tile)') % name)

    def finish_turn_fail (self, id, name):
	self.message(_('%s failed to finish turn') % name)

    def message (self, message):
	text_buffer = self.xml.get_widget('chat_text').get_buffer()
	text_buffer.insert(text_buffer.get_end_iter(), '%s\n' % message)
	vadj = self.xml.get_widget('chat_scroll').get_vadjustment()
	vadj.set_value(vadj.upper)

    def my_message (self, message):
	self.xml.get_widget('message_label').set_markup(
		'<big><b> %s </b></big>' % xml.sax.saxutils.escape(message))
    def chat (self, name, chat):
	text_buffer = self.xml.get_widget('chat_text').get_buffer()
	text_buffer.insert(text_buffer.get_end_iter(),
		'%s: %s\n' % (name, chat))
	vadj = self.xml.get_widget('chat_scroll').get_vadjustment()
	vadj.set_value(vadj.upper)

    def on_deck_window_delete_event (self, widget, event):
	self.hide_deck()
	return True

    def on_deck_menu_toggled (self, menuitem):
	if menuitem.get_active():
	    self.xml.get_widget('deck_window').show()
	else:
	    self.xml.get_widget('deck_window').hide()

    def on_control_window_delete_event (self, widget, event):
	self.hide_control()
	return True

    def on_control_menu_toggled (self, menuitem):
	if menuitem.get_active():
	    self.xml.get_widget('control_window').show()
	else:
	    self.xml.get_widget('control_window').hide()
	
    def on_score_menu_activate (self, menuitem):
	self.xml.get_widget('score_dialog').run()

    def on_score_dialog_response (self, dialog, response_id):
	dialog.hide()
